


function PaletteManagerViewModel (viewId, colorChangeDispatcher)
{
    
    var self = this;
    
    self.viewID = viewId;
    self.view = document.getElementById(viewId);
    self.paletteTabList = getDescendantById(self.view, "PaletteTabs");
    self.paletteTabContentsDiv = getDescendantById(self.view, "PaletteTabContents");
    
    // passed to palettes for communicating color changes
    self.colorChangeDispatcher = colorChangeDispatcher;
    
    self.savePaletteButton = getDescendantById(self.view, "savePaletteButton");
    self.openPaletteInput = getDescendantById(self.view, "openPaletteInput");
    
    self.palettes = new Array();
    
    self.currentTab = null;
    
    
    
    
    
    self.openPalette = function(event)
    {
        
        var file = event.target.files[0];
        var fileReader = new FileReader();
        fileReader.onload = function(event){self.createPaletteFromXml(file, event.target.result);};
        
        // set cursor to wait... doesn't work while Javascript runs...
        document.body.style.cursor = "wait";
        
        fileReader.readAsText(file);
        
        // set cursor to default
        document.body.style.cursor = "default";
    }
    
    
    self.createPaletteFromXml = function(file, xmlString)
    {
        //console.log(xmlString);
        
        // create and initialize a new palette with a unique name
        var paletteDivId = "Palette_" + self.palettes.length;
        
        var fileName = file.name;
        
        // strip off file type suffix, unless '.' is first character
        if (fileName.lastIndexOf('.') > 0)
            fileName = fileName.substring(0, fileName.lastIndexOf('.'));
        
        // create a tab and content div for the palette and add to document
        self.createTab(paletteDivId, fileName, false, true);      // not active, has close button
        
        // create PalletViewModel, which will add the palette contents
        var palette = new PaletteViewModel(paletteDivId, colorChangeDispatcher);
        palette.setAutoAddColors(false);
        palette.id = paletteDivId;
        
        // get XML doc and editor for extracting palette
        var parser = new DOMParser();
        var xmlDoc = parser.parseFromString(xmlString,"text/xml");
        var editor = new XmlDocumentEditor(xmlDoc);
        var root = xmlDoc.documentElement;
        
        // initialize palette from XML doc
        palette.initializeFromXml(root, editor);
        
        // save the palette for when it's saved
        self.palettes.push(palette);
        
    }
    
    
    self.findPaletteById = function(paletteId)
    {
        for (var i = 0; i < self.palettes.length; i++)
        {
            if (self.palettes[i].id == paletteId)
                return self.palettes[i];
        }
        
        return null;
    }
    
    
    self.setCurrentTab = function(paletteId)
    {
        var currentPaletteId = paletteId;
        self.currentPalette = self.findPaletteById(currentPaletteId);
    }
    
    
    self.closeTab = function(paletteId)
    {
        // remove tab and palette from HTML doc
        var tab = getChildById(self.paletteTabList, paletteId + "Tab");
        self.paletteTabList.removeChild(tab);
        
        var paletteDiv = getChildById(self.paletteTabContentsDiv, paletteId);
        self.paletteTabContentsDiv.removeChild(paletteDiv);
        
        // remove palette from palette list
        var palette = self.findPaletteById(paletteId);
        self.palettes.splice(self.palettes.indexOf(palette), 1);
        
    }
    
    /*
    $('a[data-toggle="tab"]').on('show.bs.tab', function (e) {
        self.currentTab = e.target; // newly activated tab
        // e.relatedTarget // previous active tab
      });
      
      
        // Add a Tab
        $('#btnAddPage').click(function() 
        {
            pageNum++;
            $('#pageTab').append(
                $('<li><a href="#page' + pageNum + '">' + 'Page ' + pageNum + '<button class="close" type="button" ' + 'title="Remove this page">×</button>' + '</a></li>'));
    
            $('#pageTabContent').append(
                $('<div class="tab-pane" id="page' + pageNum +
                '">Content page' + pageNum + '</div>'));
    
            $('#page' + pageNum).tab('show');
        }
        

    // Remove a Tab

    $('#pageTab').on('click', ' li a .close', function() {
    var tabId = $(this).parents('li').children('a').attr('href');
    $(this).parents('li').remove('li');
    $(tabId).remove();
    reNumberPages();
    $('#pageTab a:first').tab('show');
    });

    */
    
    self.createTab = function(paletteDivId, paletteTitle, setActive, hasCloseButton)
    {
        // create list item for tab
        var tab = document.createElement("li");
        tab.setAttribute("class", "nav-item");
        tab.setAttribute("id", paletteDivId + "Tab");
        
        // create link to palette panel
        var link = document.createElement("a");
        link.setAttribute("data-toggle", "tab");
        link.setAttribute("role", "tab");
        link.setAttribute("paletteId", paletteDivId);
        link.setAttribute("href", "#" + paletteDivId);
        if (setActive)
            link.setAttribute("class", "nav-link flex-sm-fill text-sm-center active");
        else
            link.setAttribute("class", "nav-link flex-sm-fill text-sm-center");
        link.addEventListener("click", function() {self.setCurrentTab(paletteDivId)});
        
        // create tab title as text node, and append to link
        var tabTitle = document.createTextNode(paletteTitle);
        
        
        
        // attach to document
        self.paletteTabList.appendChild(tab);
        tab.appendChild(link);
        link.appendChild(tabTitle);
        
        if (hasCloseButton)
        {
            var closeButton = document.createElement("button");
            closeButton.setAttribute("id", paletteDivId + "CloseButton");
            closeButton.setAttribute("class", "btn close close-button");
            closeButton.setAttribute("paletteId", paletteDivId);
            closeButton.addEventListener("click", function() {self.closeTab(paletteDivId)});
            closeButton.appendChild(document.createTextNode("X"));
            link.appendChild(closeButton);
        }
        
        
        // now create tab contents div for the palette to populate
        var paletteDiv = document.createElement("div");
        paletteDiv.setAttribute("id", paletteDivId);
        paletteDiv.setAttribute("role", "tabpanel");
        if (setActive)
            paletteDiv.setAttribute("class", "tab-pane fade in show active");
        else
            paletteDiv.setAttribute("class", "tab-pane fade in");
        
        self.paletteTabContentsDiv.appendChild(paletteDiv);
        
    }
    
    
    self.saveCurrentPalette = function(event)
    {
        var xmlString = vkbeautify.xml(self.currentPalette.generateXmlString());
        var blob = new Blob([xmlString], {type: "text/plain;charset=utf-8"});
        saveAs(blob, "palette.xml");
    }
    
    
    self.savePaletteButton.addEventListener('click', self.saveCurrentPalette);
    self.openPaletteInput.addEventListener('change', self.openPalette);
    
    
    // initialization
    {
        // add the CurrentColors tab as first
        // create and initialize a new palette with a unique name
        var currentColorsPaletteDivId = "Current_Colors";
        
        // create a tab and content div for the palette and add to document
        self.createTab(currentColorsPaletteDivId, "Current Colors", true, false);    // active, no close button
        
        // create PalletViewModel, which will add the palette contents
        var currentColorsPalette = new PaletteViewModel(currentColorsPaletteDivId, colorChangeDispatcher);
        currentColorsPalette.setAutoAddColors(true);
        currentColorsPalette.id = currentColorsPaletteDivId;
        
        // save the palette
        self.palettes.push(currentColorsPalette);
        self.currentPalette = currentColorsPalette;
    }
    

}