

// constants
RosetteViewModel.DEFAULT_DISPLAY_SCALE = 96;
RosetteViewModel.DEFAULT_ROSETTE_GRID_WIDTH = 0.001;

RosetteViewModel.DISPLAY_SCALE_INCREMENT = 5;
RosetteViewModel.MAX_DISPLAY_SCALE = 720;
RosetteViewModel.MIN_DISPLAY_SCALE = 24;

function RosetteViewModel(svg, rosette, displayScale, colorChangeDispatcher)
{
	var self = this;
	
	
    // initialization
    self.svg = svg;
	self.rosette = rosette;
	
	self.rosetteFileName = "rosette.xml";
	
	// for communicating color changes
	self.colorChangeDispatcher = colorChangeDispatcher;
	
	self.units = new Units(document.getElementById("units").value);
	
	self.unitsString = ko.observable(self.units.toUnitsString());
	
	self.displayScalePercentage = ko.observable(100);
	
	self.currentColor = NamedColor.getNamedColor("Black", new Color("#000000"));
	
	self.borderRingWidth = self.rosette.tileHeight/self.rosette.evenTile.numRows;
	
	self.designChanged = false;
	
	// svg groups, one per tile element; used when drawing to remove and replace all
	// tile elements on rosette display for corresponding row and column
	self.rosetteSvgElements;
	
	self.displayScale = displayScale;
	self.drawGrid = true;
    self.gridWidth = 0.001;
    //self.gridColor = NamedColor.getNamedColor("Light gray", new Color("#F0F0F0"));
    //self.gridColor = NamedColor.getNamedColor("Black", new Color("black"));
    
    
    
    
    self.SetDisplayValues = function()
    {
        var decimalPlaces = self.units.decimalPlaces();
        
        var symmetryInput = document.getElementById("symmetry");
        symmetryInput.value = self.rosette.symmetry;
        
        var rowSelect = document.getElementById("rowSelect");
        rowSelect.value = self.rosette.getRows();
        
        var columnSelect = document.getElementById("columnSelect");
        columnSelect.value = self.rosette.getCols();
        
        var tileWidthInput = document.getElementById("tileWidth");
        tileWidthInput.value = self.units.toUnits(self.rosette.getTileWidth()).toFixed(decimalPlaces);
        
        var tileHeightInput = document.getElementById("tileHeight");
        tileHeightInput.value = self.units.toUnits(self.rosette.getTileHeight()).toFixed(decimalPlaces);
        
        var radiusInput = document.getElementById("rosetteRadius");
        radiusInput.value = self.units.toUnits(self.rosette.getRadius()).toFixed(decimalPlaces);
        
        var borderRingWidthInput = document.getElementById("borderRingWidth");
        borderRingWidthInput.value = self.units.toUnits(self.borderRingWidth).toFixed(decimalPlaces);
    }
    
    // Initialization...
    self.SetDisplayValues();
        
    
    
    self.zoom = function()
    {
    	var scaleFactor = self.displayScalePercentage()/100;
    	// self.displayScale = RosetteViewModel.DEFAULT_DISPLAY_SCALE * scaleFactor;
    	// self.drawRosette(self.displayScale);
    	self.setSvgViewbox(scaleFactor);
    }
    
    self.displayScalePercentage.subscribe(self.zoom);
    
    
    self.setSvgViewbox = function(scaleFactor)
    {
    	// locate rosette so top edge is a few pixels down from the top of the screen
    	var radius = self.rosette.getOuterRadius() * self.displayScale + 10;
    	self.svg.attr({ viewBox: -500/scaleFactor + " " + -radius + " " + 1000/scaleFactor + " " + 1000/scaleFactor });
    }
    
    self.setSvgViewbox(1);
    
    
    self.setUnits = function(event)
    {
        if (event != null)
        {
            var unitsName = event.target.value;
            self.units = new Units(unitsName);
            self.unitsString(self.units.toUnitsString());
            self.SetDisplayValues();
        }
    }
    
    
    self.setSymmetry = function(event)
    {
        if (event != null)
        {
            var symmetry = event.target.value;
        	self.rosette.setSymmetry(symmetry);
        	self.drawRosette(self.displayScale);
        	
        	self.designChanged = true;
        }
    }
    
    
    self.setRows = function(event)
    {
        if (event != null)
        {
            var rows = event.target.value;
        	self.rosette.setRows(parseInt(rows));
        	self.drawRosette(self.displayScale);
            
            self.designChanged = true;
        }
    }
    
    
    self.setColumns = function(event)
    {
        if (event != null)
        {
            var columns = event.target.value;
        	self.rosette.setCols(parseInt(columns));
        	self.drawRosette(self.displayScale);
            
            self.designChanged = true;
        }
    }

    
    self.setTileWidth = function(event)
    {
        if (event != null)
        {
            var width = event.target.value;
        	self.rosette.setTileWidth(self.units.fromUnits(parseFloat(width)));
        	self.drawRosette(self.displayScale);
            
            self.designChanged = true;
        }
    }

    
    self.setTileHeight = function(event)
    {
        if (event != null)
        {
            var height = event.target.value;
        	self.rosette.setTileHeight(self.units.fromUnits(parseFloat(height)));
        	self.drawRosette(self.displayScale);
            
            self.designChanged = true;
        }
    }

    
    self.setRadius = function(event)
    {
    	if (event != null)
    	{
    	    var radius = event.target.value;
            self.rosette.setRadius(self.units.fromUnits(parseFloat(radius)));
        	self.drawRosette(self.displayScale);
            
            self.designChanged = true;
    	}
    }

    
    self.shiftVertically = function(shift)
    {
    	self.rosette.shiftVertically(shift);
    	self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    self.shiftHorizontally = function(shift)
    {
    	self.rosette.shiftHorizontally(shift);
    	self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    self.setBorderRingWidth = function(event)
    {
        if (event != null)
        {
            var ringWidth = event.target.value;
            self.borderRingWidth = self.units.fromUnits(parseFloat(ringWidth))
        }
    }
    
    
    self.addInnerRing = function()
    {
        
        var width = self.borderRingWidth;
        self.rosette.addInnerRing(width, self.currentColor);
        self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    self.addOuterRing = function()
    {
        
        var width = self.borderRingWidth;
        self.rosette.addOuterRing(width, self.currentColor);
        self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    self.removeInnerRing = function()
    {
        self.rosette.removeInnerRing();
        self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    self.removeOuterRing = function()
    {
        self.rosette.removeOuterRing();
        self.drawRosette(self.displayScale);
        
        self.designChanged = true;
    }
    
    
    
    
    
    
    
    self.setCurrentColor = function(namedColor)
    {
        self.currentColor = namedColor;
        
        // broadcast change to others
        self.colorChangeDispatcher.broadcastChange(self.currentColor);
    }
    
    
    // colorChangeHandler method
    self.onColorChange = function(namedColor)
    {
        self.currentColor = namedColor;
    }
    
    self.colorChangeDispatcher.addChangeListener(self, self.onColorChange);
    
    
	
    
    
    self.coloring = false;
    self.copying = false;
    self.savedCurrentColor = self.currentColor;
    
    
    self.copyColor = function()
    {
    	self.copying = true;
    }
    
    
    
    self.mouseDownHandler = function(event)
    {
    	
    	// determine if we should start coloring with current color or erasing with back color
    	if ((event != null))
		{
    		// get tile element corresponding to click
			var svg = document.getElementById("svgWindow");
	    	
	    	var pt = svg.createSVGPoint();  // Created once for document
	
	    	pt.x = event.clientX;
	    	pt.y = event.clientY;
	
		    // The cursor point, translated into svg coordinates
		    var svgPoint =  pt.matrixTransform(svg.getScreenCTM().inverse());
		    
		    //console.log("event point in screen coords: (" + pt.x + ", " + pt.y + ")");
		    //console.log("event point in svg coords:    (" + svgPoint.x + ", " + svgPoint.y + ")");
	    
		    // now see which tile element it corresponds to
		    var rosettePoint = new RealPoint(svgPoint.x / self.displayScale, svgPoint.y / self.displayScale)
		    //console.log("event point in rosette coords:    (" + rosettePoint.x + ", " + rosettePoint.y + ")");
		    
		    var rowColArray = self.findEvenTileRowCol(rosettePoint);
		    if (rowColArray != null)
			{  	
			    // we're in the rosette; set to color unless we're copying a color
		    	var row = rowColArray[0];
			    var column = rowColArray[1];
			    
			    var tileColor = self.rosette.getEvenElementColor(row, column);
			    
			    if (self.copying)
			    {
			    	
			    	// set current color to tile element color
			    	self.currentColor = tileColor;
			    	
			    	// broadcast to listeners
			    	self.colorChangeDispatcher.broadcastChange(self.currentColor);
			    }
			    else
			    {
				    // we're coloring; set true so mouse move handler will continue coloring
			    	self.coloring = true;
			    	
			    	// save current color in case we decide to erase
			    	self.savedCurrentColor = self.currentColor;
			    	
			    	// but see if we should be coloring or erasing
				    if (tileColor == self.currentColor)
				    {
				    	// we clicked on an element with the current color; change currentColor to backColor so we erase
				    	self.currentColor = self.rosette.backColor;
				    }
				    
				    // color the tile element
			    	self.rosette.setTileElementColor(row, column, self.currentColor);
			    	self.drawRosetteTileElement(row, column, self.displayScale);
			    	
			    	self.designChanged = true;
			    }
		    	
			}
		}
    }
    
    
    self.mouseUpHandler = function()
    {
    	// if coloring, restore current color in case we were erasing
    	if (self.coloring)
    		self.currentColor = self.savedCurrentColor;
    	
    	// stop coloring or copying
    	self.coloring = false;
    	self.copying = false;
    }
    
    
    self.mouseMoveHandler = function(event)
    {
    	if ( (event != null) && (self.coloring == true) )
		{
    		// get tile element corresponding to click
			var svg = document.getElementById("svgWindow");
	    	
	    	var pt = svg.createSVGPoint();  // Created once for document
	
	    	pt.x = event.clientX;
	    	pt.y = event.clientY;
	
		    // The cursor point, translated into svg coordinates
		    var svgPoint =  pt.matrixTransform(svg.getScreenCTM().inverse());
		    
		    //console.log("event point in screen coords: (" + pt.x + ", " + pt.y + ")");
		    //console.log("event point in svg coords:    (" + svgPoint.x + ", " + svgPoint.y + ")");
	    
		    // now see which tile element it corresponds to
		    var rosettePoint = new RealPoint(svgPoint.x / self.displayScale, svgPoint.y / self.displayScale)
		    //console.log("event point in rosette coords:    (" + rosettePoint.x + ", " + rosettePoint.y + ")");
		    
		    var rowColArray = self.findEvenTileRowCol(rosettePoint);
		    if (rowColArray != null)
			{  	
			    var row = rowColArray[0];
			    var column = rowColArray[1];
			    
			    var tileColor = self.rosette.getEvenElementColor(row, column);
			    
			    // color the tile element
		    	if (tileColor != self.currentColor)
			    {
			    	self.rosette.setTileElementColor(row, column, self.currentColor);
			    	self.drawRosetteTileElement(row, column, self.displayScale);
			    }
			    
			}
		
		}
    }
    
    
    var svgWindow = document.getElementById("svgWindow");
    svgWindow.addEventListener("mousedown", self.mouseDownHandler);
    svgWindow.addEventListener('mouseup', self.mouseUpHandler);
    svgWindow.addEventListener('mousemove', self.mouseMoveHandler);
    
    
    
	self.setTileElement = function(event)
	{
		if (event != null)
		{
			
			// get tile element corresponding to click
			var svg = document.getElementById("svgWindow");
	    	
	    	var pt = svg.createSVGPoint();  // Created once for document
	
	    	pt.x = event.clientX;
	    	pt.y = event.clientY;
	
		    // The cursor point, translated into svg coordinates
		    var svgPoint =  pt.matrixTransform(svg.getScreenCTM().inverse());
		    
		    //console.log("event point in screen coords: (" + pt.x + ", " + pt.y + ")");
		    //console.log("event point in svg coords:    (" + svgPoint.x + ", " + svgPoint.y + ")");
	    
		    // now see which tile element it corresponds to
		    var rosettePoint = new RealPoint(svgPoint.x / self.displayScale, svgPoint.y / self.displayScale)
		    //console.log("event point in rosette coords:    (" + rosettePoint.x + ", " + rosettePoint.y + ")");
		    
		    var rowColArray = self.findEvenTileRowCol(rosettePoint);
		    if (rowColArray != null)
			{  	
			    var row = rowColArray[0];
			    var column = rowColArray[1];
			    
			    self.rosette.setTileElementColor(row, column, self.currentColor());
			    
			    self.drawRosetteTileElement(row, column, self.displayScale);
			}
		}
	}
	
	
	
	self.findEvenTileRowCol = function(rosettePoint)
	{
		var row, column;
		
		var numRows = self.rosette.getRows();
		var numCols = self.rosette.getCols();
		
		var tileHeight = self.rosette.getTileHeight();
		var tileWidth = self.rosette.getTileWidth();
		
		var rIn = self.rosette.getRadius();
		var rOut = rIn + tileHeight;
		
		var x = rosettePoint.x;
		var y = rosettePoint.y;
		
		// row is easy - just need distance from center
		var r = Math.sqrt(x*x + y*y);
		
		if ((r < rIn) || (r > rOut))
			return null;
		
		var rowWidth = (rOut - rIn) / numRows;
		row = Math.floor((rOut - r) / rowWidth);
		
		// column is trickier - need angles
		var columnWidthAtEventRadius = tileWidth/numCols;
		var tileAngleExtent = tileWidth / rOut;
		var columnAngleExtent = (tileWidth / numCols) / rOut;
		
		// 0 angle is straight up, increases clockwise; so want arctan(x/(-y))
		var angle;
		if (y != 0)
		{
			angle = Math.atan(x/(-y));
		}
		else
		{
			angle = Math.PI/2;
		}
		
		// gives result between -PI/2 and PI/2, so adjust accordingly
		if (y > 0)
		{
			if (x <= 0)
				angle -= Math.PI;
			else
				angle += Math.PI;	
		}
		
		//console.log("Angle: " + angle);
		
		column = Math.floor(angle / columnAngleExtent);
		
		//console.log("Column: " + column);
		
		// need to remove multiples, and determine of we're on an odd or even tile
		var tileNum = Math.floor(column / numCols);
		
		//console.log("tileNum: " + tileNum);
		
		column = column - (tileNum * numCols);
		
		// now see if we're on an odd or even tile, and adjust row and column to represent the
		// position on an even tile
		if (tileNum % 2 != 0)
		{
			// need to get corresponding even tile row/column, which depends on symmetry
			// set color in appropriate element of odd tile
			switch (self.rosette.symmetry)
	        {
	            case Rosette.BOOK_MATCH:
	            {
	                column = numCols - column - 1;
	                break;
	            }
	            case Rosette.INVERSION:
	            {
	                row = numRows - row - 1;
	                break;
	            }
	            case Rosette.BOOK_INVERSION:
	            {
	                row = numRows - row - 1;
	                column = numCols - column - 1;
	                break;
	            }
	            case Rosette.SLIP_MATCH:
	            default:
	            {
	                // no alteration needed
	                break;
	            }
	        }
		}
		
		//console.log("Row and column: [" + row + ", " + column + "]");
		rowColArray = [row,column];
		return rowColArray;
		
	}
	
	
	
	
	
	
	
	self.drawRosetteTileElement = function(row, column, displayScale)
	{
		
		var pi = Math.PI;
		var numTiles;
		var numRows, numCols;
		var tileHeight, tileWidth;
		var scaledRowHeight, scaledTileHeight;
		var rIn, rOut, theta;
		var u, v, w, z, v1, v2, w1, w2;
		var elementColor;
		
		// just set color attribute of corresponding group - nice! Doesn't get much easier!
		var elementColor = self.rosette.getEvenElementColor(row,column);
		var group = self.rosetteSvgElements[row][column];
		group.attr({fill: elementColor.color.toHexString()});
	}
	
	
	
	/**
	 * Create SVG version of rosette in SVG window
	 */
	self.drawRosette = function(displayScale)
	{
		
		var pi = Math.PI;
		var numTiles;
		var numRows, numCols;
		var tileHeight, tileWidth;
		var scaledRowHeight, scaledTileHeight;
		var rIn, rOut, theta;
		var u, v, w, v1, v2, w1, w2;
		var elementColor;
		
		// spinner while waiting
		var spinner = document.getElementById("waitSpinner");
		spinner.style.visibility='visible';
		
		// first clear any existing rosette
		svg.clear();
		
		// set viewbox
		//self.setSvgViewbox();

		numRows = self.rosette.getRows();
		numCols = self.rosette.getCols();
		
		// create new SVG groups for tile elements
		self.rosetteSvgElements = new Array(numRows);
	    for (var row = 0; row < numRows; row++)
	    {
	        self.rosetteSvgElements[row] = new Array(numCols);
	    	for (var column = 0; column < numCols; column++)
	        {
	        	self.rosetteSvgElements[row][column] = svg.group();
	        }
	    }

		tileHeight = self.rosette.getTileHeight();
		tileWidth = self.rosette.getTileWidth();

		scaledTileHeight = tileHeight * displayScale;
		scaledRowHeight = scaledTileHeight / numRows;
		// height of tile and row, in fractional pixels

		rIn = self.rosette.radius * displayScale;
		rOut = (self.rosette.radius + tileHeight) * displayScale;
		// outer rosette radius, in fractional pixels
		
		u = new RealPoint(0, -rOut);
		// vector to outer radius at 12 o'clock

		numTiles = Math.floor(pi * rOut / (tileWidth * displayScale));
		// number of tiles which fit into half of rosette, rounded down

		theta = (tileWidth * displayScale / numCols) / rOut;
		// angle subtended by one column-width of tile

		for(var i = -numTiles; i < numTiles; ++i)
		{
			w = u.rotate(i*numCols*theta);
			for (var column = 0; column < numCols; ++column)
			{
				v = w.duplicate();
				w = v.rotate(theta);
				// v2 = v.duplicate();
				// w2 = w.duplicate();
				for (var row = 0; row < numRows; ++ row)
				{
					if (i % 2 == 0)	// even tile
						elementColor = self.rosette.getEvenElementColor(row,column);
					else			// odd tile
						elementColor = self.rosette.getOddElementColor(row,column);
					
					
					v1 = v.scalarMult(1 - row * scaledRowHeight / rOut);
					v2 = v.scalarMult(1 - (row+1) * scaledRowHeight / rOut);
					w1 = w.scalarMult(1 - row * scaledRowHeight / rOut);
					w2 = w.scalarMult(1 - (row+1) * scaledRowHeight / rOut);
					var xCoords = [v1.x, w1.x, w2.x, v2.x];
					var yCoords = [-v1.y, -w1.y, -w2.y, -v2.y];
					
					
					// adjust row and column for odd tile
                    var adjustedRowColumn = self.rosette.getAdjustedRowColumn(i, row, column);
                    var adjustedRow = adjustedRowColumn.row;
                    var adjustedColumn = adjustedRowColumn.column;
                    
                    var group = self.rosetteSvgElements[adjustedRow][adjustedColumn];
					
					var poly = group.polygon(v1.x, v1.y, w1.x, w1.y, w2.x, w2.y, v2.x, v2.y);
					
					
					// use the group to set attributes; only need to do once for tile 0
					if (i == 0)
					{
    					group.attr({fill: elementColor.color.toHexString()});
    					
    					if (self.drawGrid)
    					{
    					    group.attr({stroke: self.rosette.gridColor.color.toHexString(), strokeWidth: self.gridWidth*displayScale});
    					}
					}
					
					
					
				}
			}
		}
		

		// draw inner and outer rings
        var innerRings = self.rosette.innerRings;
        var innerRadius = rOut-scaledTileHeight;
        for (var i = 0; i < innerRings.length; i++)
        {
            // go from inner radius inward
            var ring = innerRings[i];
            var strokeWidth = ring.width * displayScale;
            var strokeColor = ring.namedColor.color.toHexString();
            
            // stroke width is centered along path, so offset by half stroke width
            innerRadius -= (ring.width/2 * displayScale);
            var circle = svg.circle(0, 0, innerRadius);
            circle.attr({stroke: strokeColor, strokeWidth: strokeWidth, fill: "none"});

            innerRadius -= (ring.width/2 * displayScale);
        }
        
        var outerRings = self.rosette.outerRings;
        var outerRadius = rOut;
        for (var i = 0; i < outerRings.length; i++)
        {
            // go from outer radius inward
            var ring = outerRings[i];
            var strokeWidth = ring.width * displayScale;
            var strokeColor = ring.namedColor.color.toHexString();
            
            // stroke width is centered along path, so offset by half stroke width
            outerRadius += (ring.width/2 * displayScale);
            var circle = svg.circle(0, 0, outerRadius);
            circle.attr({stroke: strokeColor, strokeWidth: strokeWidth, fill: "none"});

            outerRadius += (ring.width/2 * displayScale);
        }
        
		// outline the rosette; helps when there are no rings
		svg.circle(0, 0, innerRadius).attr({fill: "none", stroke: "black", strokeWidth: 1});
		svg.circle(0, 0, outerRadius).attr({fill: "none", stroke: "black", strokeWidth: 1});
        
        
		// draw an "x" at the center point
		var lineLength = 0.1 * displayScale;
		svg.line(-lineLength, 0, lineLength, 0).attr({stroke: "black", strokeWidth: self.gridWidth*displayScale});
		svg.line(0, -lineLength, 0, lineLength).attr({stroke: "black", strokeWidth: self.gridWidth*displayScale});
	
		spinner.style.visibility='hidden';
	}
    
    
    self.writeXML = function(/*Element*/ parent, /*XmlDocumentEditor*/ editor)
    {
        self.rosette.writeXML(parent, editor);
    }
    
    
    self.generateXmlString = function()
    {
        // create DOM document, and populate it
        var xmlDoc = new Document();
        var docHeaderComment = "Rosette data generated with RosetteDesigner Java application";
        xmlDoc.createComment(docHeaderComment);
        
        // add the standard XML processing instruction <?xml version="1.0" encoding="UTF-8"?>
        const processingInstruction = xmlDoc.createProcessingInstruction('xml', 'version="1.0" encoding="UTF-8"');
        xmlDoc.appendChild(processingInstruction);
        
        // create a "root" element for the document
        var docElementName = "rosette_designer_document"
        var docElement = xmlDoc.createElement(docElementName);
        xmlDoc.appendChild(docElement);
        
        // just a convenience wrapper that provides methods for simultaneously creating, assigning value to, and appending Elements
        var editor = new XmlDocumentEditor(xmlDoc);
        
        
        // write out rosette information to document
        self.writeXML(docElement, editor);
        
        
        // return the string contents of the doc
        var serializer = new XMLSerializer();
        var xmlString = serializer.serializeToString(xmlDoc);

        //console.log(xmlString);
        
        return xmlString;
    
    }
	
	
	self.setRosetteFromXml = function(fileName, xmlString)
	{

        self.rosetteFileName = fileName;
	    
	    //console.log(xmlString);
		var parser = new DOMParser();
		var xmlDoc = parser.parseFromString(xmlString,"text/xml");

		var editor = new XmlDocumentEditor(xmlDoc);
		var root = xmlDoc.documentElement;
		
		self.rosette.initializeFromXml(root, editor);
		
		self.borderRingWidth = self.rosette.tileHeight/self.rosette.evenTile.numRows;
		
		// update display
		self.SetDisplayValues(self.rosette);
		
		// broadcast colors to listeners
		self.broadcastRosetteColors();
		
		self.drawRosette(self.displayScale);
        
        self.designChanged = false;
	}
	
	
	self.broadcastRosetteColors = function()
	{
	    // run through list of colors in rosette and broadcast to listeners
	    
	                
	        var colors = new Array();
	        
	        var numRows = self.rosette.getRows();
	        var numCols = self.rosette.getCols();
	        
	        for(var i = 0; i < numRows; i++)
	        {
	            for(var j = 0; j < numCols; j++)
	            {
	                var color = self.rosette.getElementColor(i, j);
	                var index = colors.indexOf(color);
	                
	                if (index < 0)
	                {
	                    // didn't find; add
	                    colors.push(color);
	                }
	            }
	        }
	        
	        // now broadcast colors in list
	        for (var i = 0; i < colors.length; i++)
	        {
	            self.colorChangeDispatcher.broadcastChange(colors[i]);
	        }
	        
	}
	
	
	
	
	self.checkSave = function(yesCallback, noCallback, arg)
	{
	    bootbox.confirm(
        {
            message: "Design has unsaved changes; save first?",
            buttons: 
            {
                confirm: 
                {
                    label: 'Yes',
                    className: 'btn-primary'
                },
                
                cancel: 
                {
                    label: 'No',
                    className: 'btn-primary'
                }
            },
            onEscape: false,        // can't cancel with escape key
            closeButton: false,     // no close box - must select yes or no
            callback:   function(saveChanges)
                        {
                            if (saveChanges)
                            {
                                yesCallback(arg);
                            }
                            else
                            {
                                noCallback(arg);
                            }
                        }
                
        } );
	}
	
	
	
	self.openRosetteTrigger = function(event)
    {
        // first see if want to save current design
        if (self.designChanged)
        {
            self.checkSave(self.saveRosette, function() {document.getElementById("openRosette").click();} );
        }
        else
        {
            // trigger the open operation directly
            document.getElementById("openRosette").click();
        }
        
    };
	
	var openRosetteElement = document.getElementById('openRosetteTrigger');
    openRosetteElement.addEventListener('click', self.openRosetteTrigger);
	
	
	self.openRosette = function(event)
	{
	    var file = event.target.files[0];
        var fileReader = new FileReader();
        fileReader.onload = function(event){self.setRosetteFromXml(file.name, event.target.result);};
        
        fileReader.readAsText(file);
	}
	
	var openRosetteElement = document.getElementById('openRosette');
    openRosetteElement.addEventListener('change', self.openRosette);
    

	self.saveRosette = function()
	{
	    var xmlString = vkbeautify.xml(self.generateXmlString());
	    var blob = new Blob([xmlString], {type: "text/plain;charset=utf-8"});
	    saveAs(blob, self.rosetteFileName);
	    
	    self.designChanged = false;
	}
	
	var saveRosetteElement = document.getElementById('saveRosette');
    saveRosetteElement.addEventListener('click', self.saveRosette);
            
    
    self.saveRosetteAsPDF = async function(event)
    {
        var rosettePDFWriter = new RosettePDFWriter(self.rosette, self.drawGrid, self.gridWidth, self.rosette.gridColor, self.units);
        var blob = await rosettePDFWriter.getPDFBlob(self.rosetteFileName);
        
        var rosettePDFFileName = self.rosetteFileName;
        // strip off ".xml" suffix, unless '.' is first character
        if (rosettePDFFileName.lastIndexOf('.') > 0)
            rosettePDFFileName = rosettePDFFileName.substring(0, rosettePDFFileName.lastIndexOf('.'));
        rosettePDFFileName += ".pdf";
        
        saveAs(blob, rosettePDFFileName);
    }
    
    var saveRosetteAsPDFElement = document.getElementById('saveRosetteAsPDF');
    saveRosetteAsPDFElement.addEventListener('click', self.saveRosetteAsPDF);
    
    
    
    self.saveRosetteAsPNG = function()
    {
        // set viewbox to just contain rosette
        var radius = self.rosette.getOuterRadius() * self.displayScale + 10;
        //self.svg.attr({ viewBox: -500 + " " + -radius + " " + 1000 + " " + 1000 });
        
        saveSvgAsPng(document.getElementById("svgWindow"), "diagram.png", {top: -radius, left: -radius, height: 2*radius, width:2*radius});
    }
    
    var saveRosetteAsPNGElement = document.getElementById('saveRosetteAsPNG');
    saveRosetteAsPNGElement.addEventListener('click', self.saveRosetteAsPNG);
    
    
    self.clearRosette = function()
    {
        self.rosette.clear();
        self.drawRosette(self.displayScale);
        
        self.rosetteFileName = "rosette.xml";
        
        self.designChanged = false;
    }
    
    
    self.newRosette = function(event)
    {
        // first see if want to save current design
        if (self.designChanged)
        {
            self.checkSave(self.saveRosette, self.clearRosette);
        }
        else
        {
            // clear the rosette directly
            self.clearRosette();
        }
    }
    
    var newRosetteElement = document.getElementById('newRosette');
    newRosetteElement.addEventListener('click', self.newRosette);

    
    
    
    
    
    /*
     * 
     
    self.xmlURL = "";
    
    self.saveRosette = function()
    {
        var xmlString = vkbeautify.xml(self.generateXmlString());
        var xmlBlob = new Blob([xmlString], {type:'text/plain'});
        self.xmlURL = URL.createObjectURL(xmlBlob);
        downloads.onChanged.addListener(saveRosetteCleanup);
        
        var downloadItem = browser.downloads.download({url : self.xmlURL, filename : self.rosetteFileName});
        
        self.rosetteFileName = downloadItem.filename;
        
    }
    
    self.saveRosetteCleanup = function(downloadDelta)
    {
        if (delta.state && delta.state.current === "complete")
        {
            URL.revokeObjectURL(self.xmlURL);
        }
    } 
     
    
    self.addColorChangeListener = function(listener)
    {
        self.listeners.push(listener);
    }
    
    
    self.removeColorChangeListener = function(listener)
    {
        //self.listeners.push(listener);
    }
    
    
    self.broadcastColorChange = function()
    {
        for (var i = 0; i < self.listeners.length; i++)
        {
            self.listeners[i].onColorChange(self.currentColor);
        }
    }
    */
     
     
    /*
     
    self.colorChangeHandler = function()
    {
        var color = new Color(self.currentColorView());
        var name = self.currentColorNameView();
        self.currentColor = NamedColor.getNamedColor(name, color);
        
        self.broadcastColorChange(self.currentColor);
    }
    
    self.currentColorView.subscribe(self.colorChangeHandler);
    self.currentColorNameView.subscribe(self.colorChangeHandler);
    
    
    self.colorChangeHandler = function(event)
    {
        if (event != null)
        {
            // event will be color from color selector
            var color = new Color(event.target.value);
            
            // get current color name from UI
            var name = document.getElementById("currentColorName").value;
            
            self.currentColor = NamedColor.getNamedColor(name, color);
        }
    }
    
    
    self.colorNameChangeHandler = function(event)
    {
        if (event != null)
        {
            // event will be name from color name input box
            var name = event.target.value;
            
            self.currentColor.name = name;
        }
    }
    
    
    
    self.oldDrawRosetteTileElement = function(row, column, displayScale)
    {
        
        var pi = Math.PI;
        var numTiles;
        var numRows, numCols;
        var tileHeight, tileWidth;
        var scaledRowHeight, scaledTileHeight;
        var rIn, rOut, theta;
        var u, v, w, z, v1, v2, w1, w2;
        var elementColor;
        
        // clear svg group corresponding to tile element
        self.rosetteSvgElements[row][column].clear();

        numRows = self.rosette.getRows();
        numCols = self.rosette.getCols();

        tileHeight = self.rosette.getTileHeight();
        tileWidth = self.rosette.getTileWidth();

        scaledTileHeight = tileHeight * displayScale;
        scaledRowHeight = scaledTileHeight / numRows;
        // height of tile and row, in fractional pixels

        rIn = self.rosette.radius * displayScale;
        rOut = (self.rosette.radius + tileHeight) * displayScale;
        // outer rosette radius, in fractional pixels
        
        u = new RealPoint(0, -rOut);
        // vector to outer radius at 12 o'clock

        numTiles = Math.floor(pi * rOut / (tileWidth * displayScale));
        // number of tiles which fit into half of rosette, rounded down

        theta = (tileWidth * displayScale / numCols) / rOut;
        // angle subtended by one column-width of tile

        for(var i = -numTiles; i < numTiles; ++i)
        {
            z = u.rotate(i*numCols*theta);
            
            elementColor = self.rosette.getEvenElementColor(row,column);
            
            // adjust row and column for odd tile
            var adjustedRowColumn = self.rosette.getAdjustedRowColumn(i, row, column);
            var adjustedRow = adjustedRowColumn.row;
            var adjustedColumn = adjustedRowColumn.column;
            
            v = z.rotate(theta * adjustedColumn);
            w = z.rotate(theta * (adjustedColumn + 1));
            // v2 = v.duplicate();
            // w2 = w.duplicate();
            
            v1 = v.scalarMult(1 - adjustedRow * scaledRowHeight / rOut);
            v2 = v.scalarMult(1 - (adjustedRow+1) * scaledRowHeight / rOut);
            w1 = w.scalarMult(1 - adjustedRow * scaledRowHeight / rOut);
            w2 = w.scalarMult(1 - (adjustedRow+1) * scaledRowHeight / rOut);
            var xCoords = [v1.x, w1.x, w2.x, v2.x];
            var yCoords = [-v1.y, -w1.y, -w2.y, -v2.y];

            
            //g.fillPolygon(xCoords, yCoords, 4);
            var poly = svg.polygon(v1.x, v1.y, w1.x, w1.y, w2.x, w2.y, v2.x, v2.y);
            poly.attr({fill: elementColor.color.toHexString()});
            
            self.rosetteSvgElements[row][column].add(poly);
            
            if (self.drawGrid)
            {
                poly.attr({stroke: self.rosette.gridColor.color.toHexString(), strokeWidth: self.gridWidth*displayScale});
            }
            
        }
    
    }
    
    
    
    */
    
	
}
	